@@ -1,6 +1,6 @@
 /*
     Scan and print all the PC/SC devices available
-    Copyright (C) 2001-2022  Ludovic Rousseau <ludovic.rousseau@free.fr>
+    Copyright (C) 2001-2026  Ludovic Rousseau <ludovic.rousseau@free.fr>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -64,7 +64,7 @@ const char *pcsc_stringify_error(DWORD rv)
 		NULL))
 	{
 		/* if FormatMessage() failed */
-		snprintf(buffer, sizeof(buffer), "ERROR: 0x%08X\n", rv);
+		snprintf(buffer, sizeof(buffer), "ERROR: 0x%08lX\n", rv);
 	}
 
 	return buffer;
@@ -87,18 +87,16 @@ do { \
 
 static void usage(const char *pname)
 {
-	printf("%s usage:\n\n\t%s [ -h | -V | -n | -r | -c | -s | -t secs | -d | -p]\n\n", pname, pname);
-	printf("\t\t  -h : this help\n");
-	printf("\t\t  -V : print version number\n");
-	printf("\t\t  -n : no ATR analysis\n");
-	printf("\t\t  -r : only lists readers\n");
-	printf("\t\t  -c : only lists cards\n");
-	printf("\t\t  -s : stress mode\n");
-	printf("\t\t  -q : quiet mode\n");
-	printf("\t\t  -v : verbose mode (default)\n");
-	printf("\t\t  -t secs : quit after secs seconds\n");
-	printf("\t\t  -d : debug mode\n");
-	printf("\t\t  -p : force use of PnP mechanism\n");
+	printf("%s usage:\n\n%s [ -h | -V | -n | -r | -c | -s | -t secs | -d | -p]\n\n", pname, pname);
+	printf("  -h : this help\n");
+	printf("  -V : print version number\n");
+	printf("  -n : no ATR analysis\n");
+	printf("  -r : only lists readers\n");
+	printf("  -c : only lists cards\n");
+	printf("  -s : stress mode\n");
+	printf("  -t secs : quit after secs seconds\n");
+	printf("  -d : debug mode\n");
+	printf("  -p : force use of PnP mechanism\n");
 	printf("\n");
 }
 
@@ -121,7 +119,6 @@ typedef struct
 	bool analyse_atr;
 	bool stress_card;
 	bool print_version;
-	bool verbose;
 	bool only_list_readers;
 	bool only_list_cards;
 	bool debug;
@@ -133,6 +130,9 @@ static options_t Options;
 
 SCARDCONTEXT hContext;
 
+pthread_mutex_t spinner_mutex;
+pthread_cond_t spinner_cond;
+
 static bool is_member(const char *  item, const char * list[])
 {
 	int i = 0;
@@ -196,7 +196,6 @@ static bool should_exit(void)
 
 typedef enum
 {
-	SpinDisabled = -2,
 	SpinStopped = -1,
 	SpinRunning = 0
 } SpinState_t;
@@ -205,8 +204,10 @@ _Atomic SpinState_t spin_state = SpinStopped;
 
 static void spin_start(void)
 {
-	spin_state = Options.verbose ? SpinRunning : SpinDisabled;
+	spin_state = SpinRunning;
 	printf("%s", cnl);
+
+	pthread_cond_signal(&spinner_cond);
 }
 
 static void spin_stop(void)
@@ -214,35 +215,37 @@ static void spin_stop(void)
 	spin_state = SpinStopped;
 
 	/* clean previous output */
-	printf("%s %s ", cub2, cub2);
-	fflush(stdout);
+	printf("%s%s                         ", cub2, cub2);
+	if (Interrupted)
+		printf("\n");
+	else
+		for (int i=0; i<8; i++)
+			printf("%s", cub3);
+
+	pthread_cond_signal(&spinner_cond);
 }
 
 static void *spin_update(void *p)
 {
 	char patterns[] = {'-', '\\', '|', '/'};
 
-	/* 100 ms wait */
-	struct timespec wait_time = {.tv_sec = 0, .tv_nsec = 1000*1000*100};
+	struct timespec ts;
+	struct timeval tv;
 
 	(void)p;
 
 again:
 	/* wait until spinning starts */
-	do
-	{
-		if (should_exit())
-		{
-			if (SpinDisabled == spin_state)
-			{
-				SCardCancel(hContext);
-				pthread_exit(NULL);
-			}
-		}
+	pthread_cond_wait(&spinner_cond, &spinner_mutex);
 
-		nanosleep(&wait_time, NULL);
-	} while (spin_state < 0);
+	printf(".  (use Ctrl-C to exit)");
+	for (int i=0; i<7; i++)
+		printf("%s", cub3);
+	fflush(stdout);
 
+	gettimeofday(&tv, NULL);
+	ts.tv_sec = tv.tv_sec;
+	ts.tv_nsec = tv.tv_usec * 1000;
 	do
 	{
 		char c = patterns[spin_state];
@@ -255,11 +258,19 @@ again:
 
 		spin_state++;
 		if (spin_state >= (int)sizeof patterns)
-			spin_state = SpinRunning;
-		printf("%s %c ", cub3, c);
+			spin_state = SpinRunning;	/* 0 */
+		printf("%s%c ", cub2, c);
 		fflush(stdout);
 
-		nanosleep(&wait_time, NULL);
+		/* add 100 ms delay */
+		ts.tv_nsec += 100 * 1000 * 1000;
+		if (ts.tv_nsec > 1000 * 1000 * 1000)
+		{
+			ts.tv_nsec -= 1000* 1000 * 1000;
+			ts.tv_sec += 1;
+		}
+
+		pthread_cond_timedwait(&spinner_cond, &spinner_mutex, &ts);
 	} while (spin_state >= SpinRunning);
 
 	goto again;
@@ -290,7 +301,6 @@ static void initialize_options(options_t *options, const char *pname)
 #endif
 	options->stress_card = false;
 	options->print_version = false;
-	options->verbose = true;
 	options->only_list_readers = false;
 	options->only_list_cards = false;
 	options->debug = false;
@@ -298,12 +308,12 @@ static void initialize_options(options_t *options, const char *pname)
 	options->maxtime = 0;
 }
 
-#define OPTIONS "Vhvrcst:qdpn"
+#define OPTIONS "Vhrcst:dpn"
 
 static void print_version(void)
 {
 	printf("PC/SC device scanner\n");
-	printf("V %s (c) 2001-2022, Ludovic Rousseau <ludovic.rousseau@free.fr>\n",
+	printf("V %s (c) 2001-2026, Ludovic Rousseau <ludovic.rousseau@free.fr>\n",
 			PACKAGE_VERSION);
 }
 
@@ -325,18 +335,12 @@ static int parse_options(int argc, char *argv[], options_t *options)
 				exit(EX_OK);
 				break;
 
-			case 'v':
-				options->verbose = true;
-				break;
-
 			case 'r':
 				options->only_list_readers = true;
-				options->verbose = false;
 				break;
 
 			case 'c':
 				options->only_list_cards = true;
-				options->verbose = false;
 				options->analyse_atr = false;
 				break;
 
@@ -348,10 +352,6 @@ static int parse_options(int argc, char *argv[], options_t *options)
 				options->maxtime = atol(optarg);
 				break;
 
-			case 'q':
-				options->verbose = false;
-				break;
-
 			case 'p':
 				options->pnp = true;
 				break;
@@ -385,12 +385,7 @@ static LONG stress(SCARDCONTEXT hContext2, const char *readerName)
 	DWORD dwActiveProtocol;
 	const SCARD_IO_REQUEST *pioSendPci;
 
-	if (Options.verbose)
-	{
-		printf("Stress card in reader: %s\n\n", readerName);
-	}
-	else
-		printf("\n");
+	printf("Stress card in reader: %s\n\n", readerName);
 	rv = SCardConnect(hContext2, readerName, SCARD_SHARE_SHARED,
 			SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1, &hCard, &dwActiveProtocol);
 	if (rv != SCARD_S_SUCCESS)
@@ -425,10 +420,10 @@ static LONG stress(SCARDCONTEXT hContext2, const char *readerName)
 
 
 #define COUNT 100
-	size_t count;
+	int count;
 	for (count=0; count<COUNT; count++)
 	{
-		printf("%sFAPDU n°: %ld\n", cpl, count);
+		printf("%sFAPDU n°: %d\n", cpl, count);
 		dwSendLength = sizeof(pbSendBuffer);
 		dwRecvLength = sizeof(pbRecvBuffer);
 		rv = SCardTransmit(hCard, pioSendPci, pbSendBuffer, dwSendLength,
@@ -542,10 +537,7 @@ int main(int argc, char *argv[])
 	{
 		exit(EX_USAGE);
 	}
-	if (Options.verbose)
-	{
-		print_version();
-	}
+	print_version();
 
 	initialize_signal_handlers();
 
@@ -558,22 +550,22 @@ int main(int argc, char *argv[])
 	rv = SCardGetStatusChange(hContext, 0, rgReaderStates, 1);
 	if (! Options.pnp && rgReaderStates[0].dwEventState & SCARD_STATE_UNKNOWN)
 	{
-		if (Options.verbose)
-		{
-			printf("%sPlug'n play reader name not supported. Using polling every %d ms.%s\n", magenta, TIMEOUT, color_end);
-		}
+		printf("%sPlug'n play reader name not supported. Using polling every %d ms.%s\n", magenta, TIMEOUT, color_end);
 	}
 	else
 	{
 		Options.pnp = true;
-		if (Options.verbose)
-		{
-			printf("%sUsing reader plug'n play mechanism%s\n", magenta, color_end);
-		}
+		printf("%sUsing reader plug'n play mechanism%s\n", magenta, color_end);
 	}
 
-	/* start spining thread */
-	rv = pthread_create(&spin_pthread, NULL, spin_update, NULL);
+	if (! Options.only_list_cards && ! Options.only_list_readers)
+	{
+		pthread_cond_init(&spinner_cond, NULL);
+		pthread_mutex_init(&spinner_mutex, NULL);
+
+		/* start spining thread */
+		rv = pthread_create(&spin_pthread, NULL, spin_update, NULL);
+	}
 
 get_readers:
 	/* free memory possibly allocated in a previous loop */
@@ -595,10 +587,7 @@ get_readers:
 	 * 2. malloc the necessary storage
 	 * 3. call with the real allocated buffer
 	 */
-	if (Options.verbose)
-	{
-		printf("%sScanning present readers...%s\n", red, color_end);
-	}
+	printf("%sScanning present readers...%s\n", red, color_end);
 	rv = SCardListReaders(hContext, NULL, NULL, &dwReaders);
 	if (rv != SCARD_E_NO_READERS_AVAILABLE)
 		test_rv("SCardListReaders", rv, hContext);
@@ -641,11 +630,8 @@ get_readers:
 			return EX_OK;
 		}
 
-		if (Options.verbose)
-		{
-			printf("%sWaiting for the first reader...%s   ", red, color_end);
-			fflush(stdout);
-		}
+		printf("%sWaiting for the first reader...%s   ", red, color_end);
+		fflush(stdout);
 
 		if (Options.pnp)
 		{
@@ -674,20 +660,18 @@ get_readers:
 				rv = SCardListReaders(hContext, NULL, NULL, &dwReaders);
 				if (SCARD_E_NO_READERS_AVAILABLE == rv)
 					rv = SCARD_S_SUCCESS;
+				/* wait before next call */
 				sleep(1);
 
 				if (should_exit())
 				{
-					printf("\n");
+					spin_stop();
 					exit(EX_OK);
 				}
 			}
 			spin_stop();
 		}
-		if (Options.verbose)
-		{
-			printf("found one\n");
-		}
+		printf("found one\n");
 		goto get_readers;
 	}
 	else
@@ -934,12 +918,16 @@ get_readers:
 		if (Options.only_list_cards)
 			break;
 
+		printf("%sInsert or remove a card or a reader...%s ", red, color_end);
+		fflush(stdout);
+
 		spin_start();
 
 		rv = SCardGetStatusChange(hContext, TIMEOUT, rgReaderStates_t,
 			nbReaders);
 
 		spin_stop();
+		printf("\n");
 
 		if (Options.debug)
 			displayChangedStatus(rgReaderStates_t, nbReaders);
@@ -950,13 +938,13 @@ get_readers:
 		goto get_readers;
 
 	/* If we get out the loop, GetStatusChange() was unsuccessful */
-	if (rv != SCARD_E_CANCELLED)
-		test_rv("SCardGetStatusChange", rv, hContext);
-	else
+	test_rv("SCardGetStatusChange", rv, hContext);
+
+	if (! Options.only_list_cards && ! Options.only_list_readers)
 	{
-		printf("%s", cub3);
-		fflush(stdout);
 		pthread_join(spin_pthread, NULL);
+		pthread_mutex_destroy(&spinner_mutex);
+		pthread_cond_destroy(&spinner_cond);
 	}
 
 	/* We try to leave things as clean as possible */
